#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License.
import unittest

from azure.functions.decorators.constants import TIMER_TRIGGER
from azure.functions.decorators.core import BindingDirection, DataType
from azure.functions.decorators.timer import TimerTrigger


class TestTimer(unittest.TestCase):
    def test_timer_trigger_valid_creation(self):
        trigger = TimerTrigger(name="req",
                               schedule="dummy_schedule",
                               data_type=DataType.UNDEFINED,
                               run_on_startup=False,
                               use_monitor=False,
                               dummy_field="dummy")

        self.assertEqual(trigger.get_binding_name(), "timerTrigger")
        self.assertEqual(trigger.get_dict_repr(), {
            "type": TIMER_TRIGGER,
            "direction": BindingDirection.IN,
            'dummyField': 'dummy',
            "name": "req",
            "dataType": DataType.UNDEFINED,
            "schedule": "dummy_schedule",
            "runOnStartup": False,
            "useMonitor": False
        })



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import unittest

from azure.functions.decorators.core import BindingDirection, DataType, \
    InputBinding, OutputBinding, Trigger, Setting


class DummyTrigger(Trigger):
    @staticmethod
    def get_binding_name() -> str:
        return "Dummy"

    def __init__(self,
                 name: str,
                 data_type: DataType = DataType.UNDEFINED,
                 **kwargs):
        super().__init__(name=name, data_type=data_type)


class DummyInputBinding(InputBinding):
    @staticmethod
    def get_binding_name() -> str:
        return "DummyInputBinding"

    def __init__(self,
                 name: str,
                 data_type: DataType = DataType.UNDEFINED,
                 **kwargs):
        super().__init__(name=name, data_type=data_type)


class DummySetting(Setting):

    def __init__(self, setting_name: str) -> None:
        super().__init__(setting_name=setting_name)


class DummyOutputBinding(OutputBinding):
    @staticmethod
    def get_binding_name() -> str:
        return "DummyOutputBinding"

    def __init__(self,
                 name: str,
                 data_type: DataType = DataType.UNDEFINED,
                 **kwargs):
        super().__init__(name=name, data_type=data_type)


class TestBindings(unittest.TestCase):
    def test_trigger_creation(self):
        test_trigger = DummyTrigger(name="dummy", data_type=DataType.UNDEFINED)

        expected_dict = {'dataType': DataType.UNDEFINED,
                         'direction': BindingDirection.IN,
                         'name': 'dummy',
                         'type': 'Dummy'}
        self.assertEqual(test_trigger.get_binding_name(), "Dummy")
        self.assertEqual(test_trigger.get_dict_repr(), expected_dict)

    def test_param_direction_unset(self):
        test_trigger = DummyTrigger(name="dummy", data_type=DataType.UNDEFINED,
                                    direction="dummy", type="hello")

        expected_dict = {'dataType': DataType.UNDEFINED,
                         'direction': BindingDirection.IN,
                         'name': 'dummy',
                         'type': 'Dummy'}
        self.assertEqual(test_trigger.get_binding_name(), "Dummy")
        self.assertEqual(test_trigger.get_dict_repr(), expected_dict)

    def test_input_creation(self):
        test_input = DummyInputBinding(name="dummy",
                                       data_type=DataType.UNDEFINED)

        expected_dict = {'dataType': DataType.UNDEFINED,
                         'direction': BindingDirection.IN,
                         'name': 'dummy',
                         'type': 'DummyInputBinding'}

        self.assertEqual(test_input.get_binding_name(), "DummyInputBinding")
        self.assertEqual(test_input.get_dict_repr(), expected_dict)

    def test_output_creation(self):
        test_output = DummyOutputBinding(name="dummy",
                                         data_type=DataType.UNDEFINED)

        expected_dict = {'dataType': DataType.UNDEFINED,
                         'direction': BindingDirection.OUT,
                         'name': 'dummy',
                         'type': 'DummyOutputBinding'}

        self.assertEqual(test_output.get_binding_name(), "DummyOutputBinding")
        self.assertEqual(test_output.get_dict_repr(), expected_dict)

    def test_supported_trigger_types_populated(self):
        for supported_trigger in Trigger.__subclasses__():
            trigger_name = supported_trigger.__name__
            if trigger_name != "GenericTrigger":
                trigger_type_name = supported_trigger.get_binding_name()
                self.assertTrue(trigger_type_name is not None,
                                f"binding_type {trigger_name} can not be "
                                f"None!")
                self.assertTrue(len(trigger_type_name) > 0,
                                f"binding_type {trigger_name} can not be "
                                f"empty str!")


class TestSettings(unittest.TestCase):

    def test_setting_creation(self):
        """
        Tests that the setting_name is set correctly
        """
        # DummySetting is a test setting that inherits from Setting
        test_setting = DummySetting(setting_name="TestSetting")
        self.assertEqual(test_setting.get_setting_name(), "TestSetting")

    def test_get_dict_repr(self):
        """
        Tests that the get_dict_repr method returns the correct dict
        when a new setting is intialized
        """

        class NewSetting(DummySetting):

            def __init__(self, name: str):
                self.name = name
                super().__init__(setting_name="TestSetting")

        test_setting = NewSetting(name="NewSetting")

        expected_dict = {'setting_name': "TestSetting", "name": "NewSetting"}

        self.assertEqual(test_setting.get_dict_repr(), expected_dict)
        self.assertEqual(test_setting.get_settings_value("name"), "NewSetting")



#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License.
from typing import Optional

from azure.functions.decorators.constants import SERVICE_BUS_TRIGGER, \
    SERVICE_BUS
from azure.functions.decorators.core import Trigger, OutputBinding, DataType, \
    Cardinality, AccessRights


class ServiceBusQueueTrigger(Trigger):
    @staticmethod
    def get_binding_name() -> str:
        return SERVICE_BUS_TRIGGER

    def __init__(self,
                 name: str,
                 connection: str,
                 queue_name: str,
                 data_type: Optional[DataType] = None,
                 access_rights: Optional[AccessRights] = None,
                 is_sessions_enabled: Optional[bool] = None,
                 cardinality: Optional[Cardinality] = None,
                 **kwargs):
        self.connection = connection
        self.queue_name = queue_name
        self.access_rights = access_rights
        self.is_sessions_enabled = is_sessions_enabled
        self.cardinality = cardinality
        super().__init__(name=name, data_type=data_type)


class ServiceBusQueueOutput(OutputBinding):
    @staticmethod
    def get_binding_name() -> str:
        return SERVICE_BUS

    def __init__(self,
                 name: str,
                 connection: str,
                 queue_name: str,
                 data_type: Optional[DataType] = None,
                 access_rights: Optional[AccessRights] = None,
                 **kwargs):
        self.connection = connection
        self.queue_name = queue_name
        self.access_rights = access_rights
        super().__init__(name=name, data_type=data_type)


class ServiceBusTopicTrigger(Trigger):
    @staticmethod
    def get_binding_name() -> str:
        return SERVICE_BUS_TRIGGER

    def __init__(self,
                 name: str,
                 connection: str,
                 topic_name: str,
                 subscription_name: str,
                 data_type: Optional[DataType] = None,
                 access_rights: Optional[AccessRights] = None,
                 is_sessions_enabled: Optional[bool] = None,
                 cardinality: Optional[Cardinality] = None,
                 **kwargs):
        self.connection = connection
        self.topic_name = topic_name
        self.subscription_name = subscription_name
        self.access_rights = access_rights
        self.is_sessions_enabled = is_sessions_enabled
        self.cardinality = cardinality
        super().__init__(name=name, data_type=data_type)


class ServiceBusTopicOutput(OutputBinding):
    @staticmethod
    def get_binding_name() -> str:
        return SERVICE_BUS

    def __init__(self,
                 name: str,
                 connection: str,
                 topic_name: str,
                 subscription_name: Optional[str] = None,
                 data_type: Optional[DataType] = None,
                 access_rights: Optional[AccessRights] = None,
                 **kwargs):
        self.connection = connection
        self.topic_name = topic_name
        self.subscription_name = subscription_name
        self.access_rights = access_rights
        super().__init__(name=name, data_type=data_type)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import datetime
import typing

from azure.functions import _abc as func_abc
from azure.functions import meta


class EventHubEvent(func_abc.EventHubEvent):
    """A concrete implementation of Event Hub message type."""

    def __init__(self, *,
                 body: bytes,
                 trigger_metadata: typing.Optional[
                     typing.Mapping[str, meta.Datum]] = None,
                 enqueued_time: typing.Optional[datetime.datetime] = None,
                 partition_key: typing.Optional[str] = None,
                 sequence_number: typing.Optional[int] = None,
                 offset: typing.Optional[str] = None,
                 iothub_metadata: typing.Optional[
                     typing.Mapping[str, str]] = None) -> None:
        self.__body = body
        self.__trigger_metadata = trigger_metadata
        self.__enqueued_time = enqueued_time
        self.__partition_key = partition_key
        self.__sequence_number = sequence_number
        self.__offset = offset
        self.__iothub_metadata = iothub_metadata

        # Cache for trigger metadata after Python object conversion
        self._trigger_metadata_pyobj: typing.Optional[
            typing.Mapping[str, typing.Any]] = None

    def get_body(self) -> bytes:
        return self.__body

    @property
    def partition_key(self) -> typing.Optional[str]:
        return self.__partition_key

    @property
    def iothub_metadata(self) -> typing.Optional[typing.Mapping[str, str]]:
        return self.__iothub_metadata

    @property
    def sequence_number(self) -> typing.Optional[int]:
        return self.__sequence_number

    @property
    def enqueued_time(self) -> typing.Optional[datetime.datetime]:
        return self.__enqueued_time

    @property
    def offset(self) -> typing.Optional[str]:
        return self.__offset

    @property
    def metadata(self) -> typing.Optional[typing.Mapping[str, typing.Any]]:
        """Getting read-only trigger metadata in a Python dictionary.

        Exposing the raw trigger_metadata to our customer. For cardinality=many
        scenarios, each event points to the common metadata of all the events.

        So when using metadata field when cardinality=many, it only needs to
        take one of the events to get all the data (e.g. events[0].metadata).

        Returns:
        --------
        typing.Mapping[str, object]
            Return the Python dictionary of trigger metadata
        """
        if self.__trigger_metadata is None:
            return None

        if self._trigger_metadata_pyobj is None:
            self._trigger_metadata_pyobj = {
                k: v.python_value for (k, v) in self.__trigger_metadata.items()
            }
        return self._trigger_metadata_pyobj

    def __repr__(self) -> str:
        return (
            f'<azure.EventHubEvent '
            f'partition_key={self.partition_key} '
            f'sequence_number={self.sequence_number} '
            f'enqueued_time={self.enqueued_time} '
            f'at 0x{id(self):0x}>'
        )



#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License.
import unittest

from azure.functions.decorators.constants import SQL_TRIGGER, SQL
from azure.functions.decorators.core import BindingDirection, DataType
from azure.functions.decorators.sql import SqlTrigger, \
    SqlInput, SqlOutput


class TestSql(unittest.TestCase):
    def test_sql_trigger_valid_creation(self):
        trigger = SqlTrigger(name="req",
                             table_name="dummy_table",
                             connection_string_setting="dummy_setting",
                             data_type=DataType.UNDEFINED,
                             dummy_field="dummy")

        self.assertEqual(trigger.get_binding_name(), "sqlTrigger")
        self.assertEqual(trigger.get_dict_repr(),
                         {"connectionStringSetting": "dummy_setting",
                          "dataType": DataType.UNDEFINED,
                          "tableName": "dummy_table",
                          "direction": BindingDirection.IN,
                          "dummyField": "dummy",
                          "name": "req",
                          "type": SQL_TRIGGER})

    def test_sql_output_valid_creation(self):
        output = SqlOutput(name="req",
                           command_text="dummy_table",
                           connection_string_setting="dummy_setting",
                           data_type=DataType.UNDEFINED,
                           dummy_field="dummy")
        self.assertEqual(output.get_binding_name(), "sql")
        self.assertEqual(output.get_dict_repr(),
                         {"commandText": "dummy_table",
                          "connectionStringSetting": "dummy_setting",
                          "dataType": DataType.UNDEFINED,
                          "direction": BindingDirection.OUT,
                          "dummyField": "dummy",
                          "name": "req",
                          "type": SQL})

    def test_sql_input_valid_creation(self):
        input = SqlInput(name="req",
                         command_text="dummy_query",
                         connection_string_setting="dummy_setting",
                         data_type=DataType.UNDEFINED,
                         dummy_field="dummy")
        self.assertEqual(input.get_binding_name(), "sql")
        self.assertEqual(input.get_dict_repr(),
                         {"commandText": "dummy_query",
                          "connectionStringSetting": "dummy_setting",
                          "commandType": "Text",
                          "dataType": DataType.UNDEFINED,
                          "direction": BindingDirection.IN,
                          "dummyField": "dummy",
                          "name": "req",
                          "type": SQL})



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

from typing import Union
from . import _abc
from importlib import import_module


# Utilities
def _serialize_custom_object(obj):
    """Serialize a user-defined object to JSON.

    This function gets called when `json.dumps` cannot serialize
    an object and returns a serializable dictionary containing enough
    metadata to recontrust the original object.

    Parameters
    ----------
    obj: Object
        The object to serialize

    Returns
    -------
    dict_obj: A serializable dictionary with enough metadata to reconstruct
              `obj`

    Exceptions
    ----------
    TypeError:
        Raise if `obj` does not contain a `to_json` attribute
    """
    # 'safety' guard: raise error if object does not
    # support serialization
    if not hasattr(obj, "to_json"):
        raise TypeError(f"class {type(obj)} does not expose a `to_json` "
                        "function")
    # Encode to json using the object's `to_json`
    obj_type = type(obj)
    return {
        "__class__": obj.__class__.__name__,
        "__module__": obj.__module__,
        "__data__": obj_type.to_json(obj)
    }


def _deserialize_custom_object(obj: dict) -> object:
    """Deserialize a user-defined object from JSON.

    Deserializes a dictionary encoding a custom object,
    if it contains class metadata suggesting that it should be
    decoded further.

    Parameters:
    ----------
    obj: dict
        Dictionary object that potentially encodes a custom class

    Returns:
    --------
    object
        Either the original `obj` dictionary or the custom object it encoded

    Exceptions
    ----------
    TypeError
        If the decoded object does not contain a `from_json` function
    """
    if ("__class__" in obj) and ("__module__" in obj) and ("__data__" in obj):
        class_name = obj.pop("__class__")
        module_name = obj.pop("__module__")
        obj_data = obj.pop("__data__")

        # Importing the clas
        module = import_module(module_name)
        class_ = getattr(module, class_name)

        if not hasattr(class_, "from_json"):
            raise TypeError(f"class {type(obj)} does not expose a `from_json` "
                            "function")

        # Initialize the object using its `from_json` deserializer
        obj = class_.from_json(obj_data)
    return obj


class OrchestrationContext(_abc.OrchestrationContext):
    """A durable function orchestration context.

    :param str body:
        The body of orchestration context json.
    """

    def __init__(self,
                 body: Union[str, bytes]) -> None:
        if isinstance(body, str):
            self.__body = body
        if isinstance(body, bytes):
            self.__body = body.decode('utf-8')

    @property
    def body(self) -> str:
        return self.__body

    def __repr__(self):
        return (
            f'<azure.OrchestrationContext '
            f'body={self.body}>'
        )

    def __str__(self):
        return self.__body


class EntityContext(_abc.OrchestrationContext):
    """A durable function entity context.

    :param str body:
        The body of orchestration context json.
    """

    def __init__(self,
                 body: Union[str, bytes]) -> None:
        if isinstance(body, str):
            self.__body = body
        if isinstance(body, bytes):
            self.__body = body.decode('utf-8')

    @property
    def body(self) -> str:
        return self.__body

    def __repr__(self):
        return (
            f'<azure.EntityContext '
            f'body={self.body}>'
        )

    def __str__(self):
        return self.__body



#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License.
import inspect
import re
from abc import ABCMeta
from enum import Enum
from json import JSONEncoder
from typing import Any, TypeVar, Optional, Union, Iterable, Type, Callable

T = TypeVar("T", bound=Enum)
SNAKE_CASE_RE = re.compile(r'^([a-zA-Z]+\d*_|_+[a-zA-Z\d])\w*$')
WORD_RE = re.compile(r'^([a-zA-Z]+\d*)$')


class StringifyEnum(Enum):
    """This class output name of enum object when printed as string."""

    def __str__(self):
        return str(self.name)


class BuildDictMeta(type):
    def __new__(mcs, name, bases, dct):
        """BuildDictMeta will apply to every binding.
        It will apply :meth:`add_to_dict` decorator to :meth:`__init__` of
        every binding class to collect list of params to include in building
        json dictionary which corresponds to function.json in legacy app.
        It will also apply :meth:`skip_none` to :meth:`get_dict_repr` to
        enable json dictionary generated for every binding has non-empty
        value fields. It is needed for enabling binding param optionality.
        """
        cls = super().__new__(mcs, name, bases, dct)
        setattr(cls, '__init__',
                cls.add_to_dict(getattr(cls, '__init__')))
        setattr(cls, 'get_dict_repr',
                cls.skip_none(getattr(cls, 'get_dict_repr')))
        return cls

    @staticmethod
    def skip_none(func):
        def wrapper(*args, **kw):
            res = func(*args, **kw)
            return BuildDictMeta.clean_nones(res)

        return wrapper

    @staticmethod
    def add_to_dict(func: Callable[..., Any]):
        def wrapper(*args, **kwargs):
            if args is None or len(args) == 0:
                raise ValueError(
                    f'{func.__name__} has no args. Please ensure func is an '
                    f'object method.')

            func(*args, **kwargs)

            self = args[0]

            init_params = list(inspect.signature(func).parameters.keys())
            init_params.extend(list(kwargs.keys()))
            for key in kwargs.keys():
                if not hasattr(self, key):
                    setattr(self, key, kwargs[key])

            setattr(self, 'init_params', init_params)

        return wrapper

    @staticmethod
    def clean_nones(value):
        """
        Recursively remove all None values from dictionaries and lists,
        and returns
        the result as a new dictionary or list.
        """
        if isinstance(value, list):
            return [BuildDictMeta.clean_nones(x) for x in value if
                    x is not None]
        elif isinstance(value, dict):
            return {
                key: BuildDictMeta.clean_nones(val)
                for key, val in value.items()
                if val is not None
            }
        else:
            return value


class ABCBuildDictMeta(ABCMeta, BuildDictMeta):
    pass


def parse_singular_param_to_enum(param: Optional[Union[T, str]],
                                 class_name: Type[T]) -> Optional[T]:
    if param is None:
        return None
    if isinstance(param, str):
        try:
            return class_name[param.upper()]
        except KeyError:
            raise KeyError(
                f"Can not parse str '{param}' to {class_name.__name__}. "
                f"Allowed values are {[e.name for e in class_name]}")

    return param


def parse_iterable_param_to_enums(
        param_values: Optional[Union[Iterable[str], Iterable[T]]],
        class_name: Type[T]) -> Optional[Iterable[T]]:
    if param_values is None:
        return None

    try:
        return [class_name[value.upper()] if isinstance(value, str) else value
                for value in param_values]
    except KeyError:
        raise KeyError(
            f"Can not parse '{param_values}' to "
            f"Optional[Iterable[{class_name.__name__}]]. "
            f"Please ensure param all list elements exist in "
            f"{[e.name for e in class_name]}")


def to_camel_case(snake_case_str: str):
    if snake_case_str is None or len(snake_case_str) == 0:
        raise ValueError(
            f"Please ensure arg name {snake_case_str} is not empty!")

    if not is_snake_case(snake_case_str) and not is_word(snake_case_str):
        raise ValueError(
            f"Please ensure {snake_case_str} is a word or snake case "
            f"string with underscore as separator.")
    words = snake_case_str.split('_')
    return words[0] + ''.join([ele.title() for ele in words[1:]])


def is_snake_case(input_string: str) -> bool:
    """
    Checks if a string is formatted as "snake case".
    A string is considered snake case when:
    - it's composed only by lowercase/uppercase letters and digits
    - it contains at least one underscore
    - it does not start with a number
    *Examples:*
    >>> is_snake_case('foo_bar_baz') # returns true
    >>> is_snake_case('foo') # returns false
    :param input_string: String to test.
    :return: True for a snake case string, false otherwise.
    """
    return SNAKE_CASE_RE.match(input_string) is not None


def is_word(input_string: str) -> bool:
    """
    Checks if a string is one word.
    A string is considered one word when:
    - it's composed only by lowercase/uppercase letters and digits
    - it does not start with a number
    *Examples:*
    >>> is_word('1foo') # returns false
    >>> is_word('foo_') # returns false
    >>> is_word('foo') # returns true
    :param input_string: String to test.
    :return: True for one word string, false otherwise.
    """
    return WORD_RE.match(input_string) is not None


class StringifyEnumJsonEncoder(JSONEncoder):
    def default(self, o):
        if isinstance(o, StringifyEnum):
            return str(o)

        return super().default(o)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
from typing import Optional, Iterable

from azure.functions.decorators.constants import HTTP_TRIGGER, HTTP_OUTPUT
from azure.functions.decorators.core import AuthLevel, Trigger, \
    OutputBinding, DataType, StringifyEnum


class HttpMethod(StringifyEnum):
    """All http methods Azure Python function supports."""
    GET = "GET"
    POST = "POST"
    DELETE = "DELETE"
    HEAD = "HEAD"
    PATCH = "PATCH"
    PUT = "PUT"
    OPTIONS = "OPTIONS"


class HttpTrigger(Trigger):
    @staticmethod
    def get_binding_name() -> str:
        return HTTP_TRIGGER

    def __init__(self,
                 name,
                 methods: Optional[Iterable[HttpMethod]] = None,
                 data_type: Optional[DataType] = None,
                 auth_level: Optional[AuthLevel] = None,
                 route: Optional[str] = None,
                 **kwargs) -> None:
        self.auth_level = auth_level
        self.methods = methods
        self.route = route
        super().__init__(name=name, data_type=data_type)


class HttpOutput(OutputBinding):
    @staticmethod
    def get_binding_name() -> str:
        return HTTP_OUTPUT

    def __init__(self,
                 name: str,
                 data_type: Optional[DataType] = None,
                 **kwargs) -> None:
        super().__init__(name=name, data_type=data_type)



#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License.
from typing import Optional

from azure.functions.decorators.core import Setting

RETRY_POLICY = "retry_policy"


class RetryPolicy(Setting):

    def __init__(self,
                 strategy: str,
                 max_retry_count: str,
                 delay_interval: Optional[str] = None,
                 minimum_interval: Optional[str] = None,
                 maximum_interval: Optional[str] = None,
                 **kwargs):
        self.strategy = strategy
        self.max_retry_count = max_retry_count
        self.delay_interval = delay_interval
        self.minimum_interval = minimum_interval
        self.maximum_interval = maximum_interval
        super().__init__(setting_name=RETRY_POLICY)



#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License.

from azure.functions.decorators.core import Setting

FUNCTION_NAME = "function_name"


class FunctionName(Setting):

    def __init__(self, function_name: str,
                 **kwargs):
        self.function_name = function_name
        super().__init__(setting_name=FUNCTION_NAME)



#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License.
from typing import Optional

from azure.functions.decorators.constants import SQL, SQL_TRIGGER
from azure.functions.decorators.core import DataType, InputBinding, \
    OutputBinding, Trigger


class SqlInput(InputBinding):
    @staticmethod
    def get_binding_name() -> str:
        return SQL

    def __init__(self,
                 name: str,
                 command_text: str,
                 connection_string_setting: str,
                 command_type: Optional[str] = 'Text',
                 parameters: Optional[str] = None,
                 data_type: Optional[DataType] = None,
                 **kwargs):
        self.command_text = command_text
        self.connection_string_setting = connection_string_setting
        self.command_type = command_type
        self.parameters = parameters
        super().__init__(name=name, data_type=data_type)


class SqlOutput(OutputBinding):
    @staticmethod
    def get_binding_name() -> str:
        return SQL

    def __init__(self,
                 name: str,
                 command_text: str,
                 connection_string_setting: str,
                 data_type: Optional[DataType] = None,
                 **kwargs):
        self.command_text = command_text
        self.connection_string_setting = connection_string_setting
        super().__init__(name=name, data_type=data_type)


class SqlTrigger(Trigger):
    @staticmethod
    def get_binding_name() -> str:
        return SQL_TRIGGER

    def __init__(self,
                 name: str,
                 table_name: str,
                 connection_string_setting: str,
                 leases_table_name: Optional[str] = None,
                 data_type: Optional[DataType] = None,
                 **kwargs):
        self.table_name = table_name
        self.connection_string_setting = connection_string_setting
        self.leases_table_name = leases_table_name
        super().__init__(name=name, data_type=data_type)



from typing import List


class CollectionBytes:
    """The CollectionBytes class is used for generating a mock
    'collection_bytes' meta.Datum in testing. The common usage of it is
    new_datum = meta.Datum(type='collection_bytes',
                           value=CollectionBytes([b'1', b'2']))
    """
    def __init__(self, data: List[bytes]):
        self.bytes = data


class CollectionString:
    """The CollectionString class is used for generating a mock
    'collection_string' meta.Datum in testing. The common usage of it is
    new_datum = meta.Datum(type='collection_string',
                           value=CollectionString(['a', 'b']))
    """
    def __init__(self, data: List[str]):
        self.string = data


class CollectionSint64:
    """The CollectionSint64 class is used for generating a mock
    'collection_sint64' meta.Datum in testing. The common usage of it is
    new_datum = meta.Datum(type='collection_sint64',
                           value=CollectionSint64([1, 2]))
    """
    def __init__(self, data: List[int]):
        self.sint64 = data



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import datetime
from typing import Optional, Dict, Any, Union

from . import _abc


class ServiceBusMessage(_abc.ServiceBusMessage):

    """A ServiceBuss message object.

    :param body:
        A string or bytes instance specifying the message body.

    :param content_type:
        An optional string specifying the content type.

    :param correlation_id:
        An optional string specifying the correlation id.

    """

    def __init__(self, *,
                 body: Optional[Union[str, bytes]] = None,
                 content_type: Optional[str] = None,
                 correlation_id: Optional[str] = None) -> None:
        self.__body = b''
        self.__content_type = content_type
        self.__correlation_id = correlation_id

        if body is not None:
            self.__set_body(body)

    @property
    def content_type(self) -> Optional[str]:
        """Optionally describes the payload of the message,
        with a descriptor following the format of RFC2045

        Returns:
        --------
        Optional[str]
            If content type is set, returns a string.
            Otherwise, returns None.
        """
        return self.__content_type

    @property
    def correlation_id(self) -> Optional[str]:
        """Enables an application to specify a context for the message for the
        purposes of correlation

        Returns:
        --------
        Optional[str]
            If correlation id set, returns a string.
            Otherwise, returns None.
        """
        return self.__correlation_id

    @property
    def dead_letter_source(self) -> Optional[str]:
        """Only set in messages that have been dead-lettered and subsequently
        auto-forwarded from the dead-letter queue to another entity.
        Indicates the entity in which the message was dead-lettered.
        This property is read-only.

        Returns:
        --------
        Optional[str]
            If dead letter source is set, returns a string.
            Otherwise, returns None.
        """
        return None

    @property
    def delivery_count(self) -> Optional[int]:
        """Number of deliveries that have been attempted for this message.
        The count is incremented when a message lock expires,
        or the message is explicitly abandoned by the receiver.
        This property is read-only.

        Returns:
        --------
        Optional[str]
            If delivery count is set, returns a string.
            Otherwise, returns None.
        """
        return None

    @property
    def enqueued_time_utc(self) -> Optional[datetime.datetime]:
        """The UTC instant at which the message has been accepted and stored
        in the entity. This value can be used as an authoritative and neutral
        arrival time indicator when the receiver does not want to trust the
        sender's clock. This property is read-only.

        Returns:
        --------
        Optional[datetime.datetime]
            If enqueued time utc is set, returns a datetime.
            Otherwise, returns None.
        """
        return None

    @property
    def expires_at_utc(self) -> Optional[datetime.datetime]:
        """The UTC instant at which the message is marked for removal and no
        longer available for retrieval from the entity due to its expiration.
        Expiry is controlled by the TimeToLive property and this property is
        computed from EnqueuedTimeUtc+TimeToLive. This property is read-only.

        Returns:
        --------
        Optional[datetime.datetime]
            If expires at utc is set, returns a datetime.
            Otherwise, returns None.
        """
        return None

    @property
    def expiration_time(self) -> Optional[datetime.datetime]:
        """(Deprecated, use expires_at_utc instead)"""
        return None

    @property
    def label(self) -> Optional[str]:
        """This property enables the application to indicate the purpose of
        the message to the receiver in a standardized fashion, similar to an
        email subject line.

        Returns:
        --------
        Optional[str]
            If label is set, returns a string.
            Otherwise, returns None.
        """
        return None

    @property
    def lock_token(self) -> Optional[str]:
        """	The lock token is a reference to the lock that is being held by
        the broker in peek-lock receive mode. The token can be used to pin the
        lock permanently through the Deferral API and, with that, take the
        message out of the regular delivery state flow.
        This property is read-only.

        Returns:
        --------
        Optional[str]
            If local token is set, returns a string.
            Otherwise, returns None.
        """
        return None

    @property
    def message_id(self) -> str:
        """The message identifier is an application-defined value that
        uniquely identifies the message and its payload.
        The identifier is a free-form string and can reflect a GUID or an
        identifier derived from the application context. If enabled, the
        duplicate detection feature identifies and removes second and further
        submissions of messages with the same MessageId.

        Returns:
        --------
        str
            The message identifier
        """
        return ""

    @property
    def partition_key(self) -> Optional[str]:
        """	For partitioned entities, setting this value enables assigning
        related messages to the same internal partition, so that submission
        sequence order is correctly recorded. The partition is chosen by a
        hash function over this value and cannot be chosen directly. For
        session-aware entities, the SessionId property overrides this value.

        Returns:
        --------
        Optional[str]
            If partition key is set, returns a string.
            Otherwise, returns None.
        """
        return None

    @property
    def reply_to(self) -> Optional[str]:
        """This optional and application-defined value is a standard way to
        express a reply path to the receiver of the message. When a sender
        expects a reply, it sets the value to the absolute or relative path
        of the queue or topic it expects the reply to be sent to.

        Returns:
        --------
        Optional[str]
            If reply to is set, returns a string.
            Otherwise, returns None.
        """
        return None

    @property
    def reply_to_session_id(self) -> Optional[str]:
        """This value augments the ReplyTo information and specifies which
        SessionId should be set for the reply when sent to the reply entity.

        Returns:
        --------
        Optional[str]
            If reply to session id is set, returns a string.
            Otherwise, returns None.
        """
        return None

    @property
    def scheduled_enqueue_time(self) -> Optional[datetime.datetime]:
        """(Deprecated, use scheduled_enqueue_time_utc instead)"""
        return None

    @property
    def scheduled_enqueue_time_utc(self) -> Optional[datetime.datetime]:
        """For messages that are only made available for retrieval after a
        delay, this property defines the UTC instant at which the message
        will be logically enqueued, sequenced, and therefore made available
        for retrieval.

        Returns:
        --------
        Optional[datetime.datetime]
            If scheduled enqueue time utc is set, returns a string.
            Otherwise, returns None.
        """
        return None

    @property
    def sequence_number(self) -> Optional[int]:
        """The sequence number is a unique 64-bit integer assigned to a message
        as it is accepted and stored by the broker and functions as its true
        identifier. For partitioned entities, the topmost 16 bits reflect the
        partition identifier. Sequence numbers monotonically increase and are
        gapless. They roll over to 0 when the 48-64 bit range is exhausted.
        This property is read-only.

        Returns:
        --------
        Optional[int]
            If sequence number is set, returns an integer.
            Otherwise, returns None.
        """
        return None

    @property
    def session_id(self) -> Optional[str]:
        """For session-aware entities, this application-defined value
        specifies the session affiliation of the message. Messages with the
        same session identifier are subject to summary locking and enable
        exact in-order processing and demultiplexing. For entities that are
        not session-aware, this value is ignored.

        Returns:
        --------
        Optional[str]
            If session id is set, returns a string.
            Otherwise, returns None.
        """
        return None

    @property
    def time_to_live(self) -> Optional[datetime.timedelta]:
        """	This value is the relative duration after which the message
        expires, starting from the instant the message has been accepted and
        stored by the broker, as captured in EnqueueTimeUtc. When not set
        explicitly, the assumed value is the DefaultTimeToLive for the
        respective queue or topic. A message-level TimeToLive value cannot
        be longer than the entity's DefaultTimeToLive setting.
        If it is longer, it is silently adjusted.

        Returns:
        --------
        Optional[datetime.timedelta]
            If time to live is set, returns a timedelta.
            Otherwise, returns None.
        """
        return None

    @property
    def to(self) -> Optional[str]:
        """	This property is reserved for future use in routing scenarios and
        currently ignored by the broker itself. Applications can use this
        value in rule-driven auto-forward chaining scenarios to indicate the
        intended logical destination of the message.

        Returns:
        --------
        Optional[str]
            If the recipient is set, returns a string.
            Otherwise, returns None.
        """
        return None

    @property
    def user_properties(self) -> Dict[str, Any]:
        """Contains user defined message properties.

        Returns:
        --------
        Dict[str, Any]:
            If user has set properties for the message, returns a dictionary.
            If nothing is set, returns an empty dictionary.
        """
        return {}

    @property
    def metadata(self) -> Optional[Dict[str, Any]]:
        """Getting read-only trigger metadata in a Python dictionary.

        Exposing the raw trigger_metadata to our customer. For cardinality=many
        scenarios, each event points to the common metadata of all the events.

        So when using metadata field when cardinality=many, it only needs to
        take one of the events to get all the data (e.g. events[0].metadata).

        Returns:
        --------
        Dict[str, object]
            Return the Python dictionary of trigger metadata
        """
        return None

    def __set_body(self, body):
        if isinstance(body, str):
            body = body.encode('utf-8')

        if not isinstance(body, (bytes, bytearray)):
            raise TypeError(
                f'response is expected to be either of '
                f'str, bytes, or bytearray, got {type(body).__name__}')

        self.__body = bytes(body)

    def get_body(self) -> bytes:
        """Return message content as bytes."""
        return self.__body



#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License.

# binding types
COSMOS_DB = "cosmosDB"
COSMOS_DB_TRIGGER = "cosmosDBTrigger"
EVENT_HUB_TRIGGER = "eventHubTrigger"
EVENT_HUB = "eventHub"
HTTP_TRIGGER = "httpTrigger"
HTTP_OUTPUT = "http"
QUEUE = "queue"
QUEUE_TRIGGER = "queueTrigger"
SERVICE_BUS = "serviceBus"
SERVICE_BUS_TRIGGER = "serviceBusTrigger"
TIMER_TRIGGER = "timerTrigger"
WARMUP_TRIGGER = "warmupTrigger"
BLOB_TRIGGER = "blobTrigger"
BLOB = "blob"
EVENT_GRID_TRIGGER = "eventGridTrigger"
EVENT_GRID = "eventGrid"
TABLE = "table"
SQL = "sql"
SQL_TRIGGER = "sqlTrigger"
DAPR_SERVICE_INVOCATION_TRIGGER = "daprServiceInvocationTrigger"
DAPR_BINDING_TRIGGER = "daprBindingTrigger"
DAPR_TOPIC_TRIGGER = "daprTopicTrigger"
DAPR_STATE = "daprState"
DAPR_SECRET = "daprSecret"
DAPR_PUBLISH = "daprPublish"
DAPR_INVOKE = "daprInvoke"
DAPR_PUBLISH = "daprPublish"
DAPR_BINDING = "daprBinding"



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import asyncio
import threading
import unittest

import azure.functions as func
from azure.functions._abc import TraceContext, RetryContext
from azure.functions._http_asgi import (
    AsgiMiddleware
)
import pytest


class MockAsgiApplication:
    response_code = 200
    response_body = b''
    response_headers = [
        [b"content-type", b"text/plain"],
    ]
    startup_called = False
    shutdown_called = False

    def __init__(self, fail_startup=False, fail_shutdown=False):
        self.fail_startup = fail_startup
        self.fail_shutdown = fail_shutdown

    async def __call__(self, scope, receive, send):
        self.received_scope = scope

        # Verify against ASGI specification
        assert scope['asgi.spec_version'] in ['2.0', '2.1']
        assert isinstance(scope['asgi.spec_version'], str)

        assert scope['asgi.version'] in ['2.0', '2.1', '2.2']
        assert isinstance(scope['asgi.version'], str)

        assert isinstance(scope['type'], str)

        if scope['type'] == 'lifespan':
            self.startup_called = True
            startup_message = await receive()
            assert startup_message['type'] == 'lifespan.startup'
            if self.fail_startup:
                if isinstance(self.fail_startup, str):
                    await send({
                        "type": "lifespan.startup.failed",
                        "message": self.fail_startup})
                else:
                    await send({"type": "lifespan.startup.failed"})
            else:
                await send({"type": "lifespan.startup.complete"})
            shutdown_message = await receive()
            assert shutdown_message['type'] == 'lifespan.shutdown'
            if self.fail_shutdown:
                if isinstance(self.fail_shutdown, str):
                    await send({
                        "type": "lifespan.shutdown.failed",
                        "message": self.fail_shutdown})
                else:
                    await send({"type": "lifespan.shutdown.failed"})
            else:
                await send({"type": "lifespan.shutdown.complete"})

            self.shutdown_called = True

        elif scope['type'] == 'http':
            assert scope['http_version'] in ['1.0', '1.1', '2']
            assert isinstance(scope['http_version'], str)

            assert scope['method'] in ['POST', 'GET', 'PUT', 'DELETE', 'PATCH']
            assert isinstance(scope['method'], str)

            assert scope['scheme'] in ['http', 'https']
            assert isinstance(scope['scheme'], str)

            assert isinstance(scope['path'], str)
            assert isinstance(scope['raw_path'], bytes)
            assert isinstance(scope['query_string'], bytes)
            assert isinstance(scope['root_path'], str)

            assert hasattr(scope['headers'], '__iter__')
            for k, v in scope['headers']:
                assert isinstance(k, bytes)
                assert isinstance(v, bytes)

            assert scope['client'] is None or hasattr(scope['client'],
                                                      '__iter__')
            if scope['client']:
                assert len(scope['client']) == 2
                assert isinstance(scope['client'][0], str)
                assert isinstance(scope['client'][1], int)

            assert scope['server'] is None or hasattr(scope['server'],
                                                      '__iter__')
            if scope['server']:
                assert len(scope['server']) == 2
                assert isinstance(scope['server'][0], str)
                assert isinstance(scope['server'][1], int)

            self.received_request = await receive()
            assert self.received_request['type'] == 'http.request'
            assert isinstance(self.received_request['body'], bytes)
            assert isinstance(self.received_request['more_body'], bool)

            await send(
                {
                    "type": "http.response.start",
                    "status": self.response_code,
                    "headers": self.response_headers,
                }
            )
            await send(
                {
                    "type": "http.response.body",
                    "body": self.response_body,
                }
            )

            self.next_request = await receive()
            assert self.next_request['type'] == 'http.disconnect'
        else:
            raise AssertionError(f"unexpected type {scope['type']}")


class TestHttpAsgiMiddleware(unittest.TestCase):
    def _generate_func_request(
            self,
            method="POST",
            url="https://function.azurewebsites.net/api/http?firstname=rt",
            headers={
                "Content-Type": "application/json",
                "x-ms-site-restricted-token": "xmsrt"
            },
            params={
                "firstname": "roger"
            },
            route_params={},
            body=b'{ "lastname": "tsang" }'
    ) -> func.HttpRequest:
        return func.HttpRequest(
            method=method,
            url=url,
            headers=headers,
            params=params,
            route_params=route_params,
            body=body
        )

    def _generate_func_context(
        self,
        invocation_id='123e4567-e89b-12d3-a456-426655440000',
        thread_local_storage=threading.local(),
        function_name='httptrigger',
        function_directory='/home/roger/wwwroot/httptrigger',
        trace_context=TraceContext,
        retry_context=RetryContext
    ) -> func.Context:
        class MockContext(func.Context):
            def __init__(self, ii, tls, fn, fd, tc, rc):
                self._invocation_id = ii
                self._thread_local_storage = tls
                self._function_name = fn
                self._function_directory = fd
                self._trace_context = tc
                self._retry_context = rc

            @property
            def invocation_id(self):
                return self._invocation_id

            @property
            def thread_local_storage(self):
                return self._thread_local_storage

            @property
            def function_name(self):
                return self._function_name

            @property
            def function_directory(self):
                return self._function_directory

            @property
            def trace_context(self):
                return self._trace_context

            @property
            def retry_context(self):
                return self._retry_context

        return MockContext(invocation_id, thread_local_storage, function_name,
                           function_directory, trace_context, retry_context)

    def test_middleware_calls_app(self):
        app = MockAsgiApplication()
        test_body = b'Hello world!'
        app.response_body = test_body
        app.response_code = 200
        req = func.HttpRequest(method='get', url='/test', body=b'')
        response = AsgiMiddleware(app).handle(req)

        # Verify asserted
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.get_body(), test_body)

    def test_middleware_calls_app_with_context(self):
        """Test if the middleware can be used by exposing the .handle method,
        specifically when the middleware is used as
        def main(req, context):
            return AsgiMiddleware(app).handle(req, context)
        """
        app = MockAsgiApplication()
        test_body = b'Hello world!'
        app.response_body = test_body
        app.response_code = 200
        req = self._generate_func_request()
        ctx = self._generate_func_context()
        response = AsgiMiddleware(app).handle(req, ctx)

        # Verify asserted
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.get_body(), test_body)

    def test_middleware_wrapper(self):
        """Test if the middleware can be used by exposing the .main property,
        specifically when the middleware is used as
        main = AsgiMiddleware(app).main
        """
        app = MockAsgiApplication()
        test_body = b'Hello world!'
        app.response_body = test_body
        app.response_code = 200
        req = self._generate_func_request()
        ctx = self._generate_func_context()

        main = AsgiMiddleware(app).main
        response = main(req, ctx)

        # Verify asserted
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.get_body(), test_body)

    def test_middleware_async_calls_app_with_context(self):
        """Test the middleware with the awaitable handle_async() method
        async def main(req, context):
            return await AsgiMiddleware(app).handle_async(req, context)
        """
        app = MockAsgiApplication()
        test_body = b'Hello world!'
        app.response_body = test_body
        app.response_code = 200
        req = self._generate_func_request()
        ctx = self._generate_func_context()
        response = asyncio.get_event_loop().run_until_complete(
            AsgiMiddleware(app).handle_async(req, ctx)
        )

        # Verify asserted
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.get_body(), test_body)

    def test_function_app_lifecycle_events(self):
        mock_app = MockAsgiApplication()
        middleware = AsgiMiddleware(mock_app)
        asyncio.get_event_loop().run_until_complete(
            middleware.notify_startup()
        )
        assert mock_app.startup_called

        asyncio.get_event_loop().run_until_complete(
            middleware.notify_shutdown()
        )
        assert mock_app.shutdown_called

    def test_function_app_lifecycle_events_with_failures(self):
        apps = [
            MockAsgiApplication(False, True),
            MockAsgiApplication(True, False),
            MockAsgiApplication(True, True),
            MockAsgiApplication("bork", False),
            MockAsgiApplication(False, "bork"),
            MockAsgiApplication("bork", "bork"),
        ]
        for mock_app in apps:
            middleware = AsgiMiddleware(mock_app)
            asyncio.get_event_loop().run_until_complete(
                middleware.notify_startup()
            )
            assert mock_app.startup_called

            asyncio.get_event_loop().run_until_complete(
                middleware.notify_shutdown()
            )
            assert mock_app.shutdown_called

    def test_calling_shutdown_without_startup_errors(self):
        mock_app = MockAsgiApplication()
        middleware = AsgiMiddleware(mock_app)
        with pytest.raises(RuntimeError):
            asyncio.get_event_loop().run_until_complete(
                middleware.notify_shutdown()
            )



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import collections
import json

from . import _abc


class Document(_abc.Document, collections.UserDict):
    """An Azure Document.

    Document objects are ``UserDict`` subclasses and behave like dicts.
    """

    @classmethod
    def from_json(cls, json_data: str) -> 'Document':
        """Create a Document from a JSON string."""
        return cls.from_dict(json.loads(json_data))

    @classmethod
    def from_dict(cls, dct: dict) -> 'Document':
        """Create a Document from a dict object."""
        return cls({k: v for k, v in dct.items()})

    def to_json(self) -> str:
        """Return the JSON representation of the document."""
        return json.dumps(dict(self))

    def to_dict(self) -> dict:
        """Return the document as a dict - directly using self would also work
        as Document is ``UserDict`` subclass and behave like dict"""
        return dict(self)

    def __getitem__(self, key):
        return collections.UserDict.__getitem__(self, key)

    def __setitem__(self, key, value):
        return collections.UserDict.__setitem__(self, key, value)

    def __repr__(self) -> str:
        return (
            f'<azure.Document at 0x{id(self):0x}>'
        )


class DocumentList(_abc.DocumentList, collections.UserList):
    "A ``UserList`` subclass containing a list of :class:`~Document` objects"
    pass



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import datetime
import json
from typing import Dict, Any, List, Union, Optional, Mapping, cast

from azure.functions import _servicebus as azf_sbus

from . import meta


class ServiceBusMessage(azf_sbus.ServiceBusMessage):
    """An HTTP response object."""

    def __init__(
            self, *,
            body: bytes,
            trigger_metadata: Optional[Mapping[str, Any]] = None,
            content_type: Optional[str] = None,
            correlation_id: Optional[str] = None,
            dead_letter_source: Optional[str] = None,
            delivery_count: Optional[int] = None,
            enqueued_time_utc: Optional[datetime.datetime] = None,
            expires_at_utc: Optional[datetime.datetime] = None,
            label: Optional[str] = None,
            lock_token: Optional[str] = None,
            message_id: str,
            partition_key: Optional[str] = None,
            reply_to: Optional[str] = None,
            reply_to_session_id: Optional[str] = None,
            scheduled_enqueue_time_utc: Optional[datetime.datetime] = None,
            sequence_number: Optional[int] = None,
            session_id: Optional[str] = None,
            time_to_live: Optional[datetime.timedelta] = None,
            to: Optional[str] = None,
            user_properties: Dict[str, object]):
        super().__init__(body=body, content_type=content_type,
                         correlation_id=correlation_id)
        self.__body = body
        self.__trigger_metadata = trigger_metadata
        self.__content_type = content_type
        self.__correlation_id = correlation_id
        self.__dead_letter_source = dead_letter_source
        self.__delivery_count = delivery_count
        self.__enqueued_time_utc = enqueued_time_utc
        self.__expires_at_utc = expires_at_utc
        self.__label = label
        self.__lock_token = lock_token
        self.__message_id = message_id
        self.__partition_key = partition_key
        self.__reply_to = reply_to
        self.__reply_to_session_id = reply_to_session_id
        self.__scheduled_enqueue_time_utc = scheduled_enqueue_time_utc
        self.__sequence_number = sequence_number
        self.__session_id = session_id
        self.__time_to_live = time_to_live
        self.__to = to
        self.__user_properties = user_properties

        # Cache for trigger metadata after Python object conversion
        self._trigger_metadata_pyobj: Optional[Dict[str, Any]] = None

    def get_body(self) -> bytes:
        return self.__body

    @property
    def content_type(self) -> Optional[str]:
        return self.__content_type

    @property
    def correlation_id(self) -> Optional[str]:
        return self.__correlation_id

    @property
    def dead_letter_source(self) -> Optional[str]:
        return self.__dead_letter_source

    @property
    def delivery_count(self) -> Optional[int]:
        return self.__delivery_count

    @property
    def enqueued_time_utc(self) -> Optional[datetime.datetime]:
        return self.__enqueued_time_utc

    @property
    def expires_at_utc(self) -> Optional[datetime.datetime]:
        return self.__expires_at_utc

    @property
    def expiration_time(self) -> Optional[datetime.datetime]:
        """(Deprecated) Use expires_at_utc instead"""
        return self.__expires_at_utc

    @property
    def force_persistence(self) -> Optional[bool]:
        # This field is not exposed to the customer in _servicebus.py interface
        # Need to confirm how gRPC pass boolean value
        raise NotImplementedError('This requires gRPC boolean support')

    @property
    def label(self) -> Optional[str]:
        return self.__label

    @property
    def lock_token(self) -> Optional[str]:
        return self.__lock_token

    @property
    def message_id(self) -> str:
        return self.__message_id

    @property
    def partition_key(self) -> Optional[str]:
        return self.__partition_key

    @property
    def reply_to(self) -> Optional[str]:
        return self.__reply_to

    @property
    def reply_to_session_id(self) -> Optional[str]:
        return self.__reply_to_session_id

    @property
    def scheduled_enqueue_time(self) -> Optional[datetime.datetime]:
        """(Deprecated) Use scheduled_enqueue_time_utc instead"""
        return self.__scheduled_enqueue_time_utc

    @property
    def scheduled_enqueue_time_utc(self) -> Optional[datetime.datetime]:
        return self.__scheduled_enqueue_time_utc

    @property
    def sequence_number(self) -> Optional[int]:
        return self.__sequence_number

    @property
    def session_id(self) -> Optional[str]:
        return self.__session_id

    @property
    def time_to_live(self) -> Optional[datetime.timedelta]:
        return self.__time_to_live

    @property
    def to(self) -> Optional[str]:
        return self.__to

    @property
    def user_properties(self) -> Dict[str, object]:
        return self.__user_properties

    @property
    def metadata(self) -> Optional[Dict[str, Any]]:
        if self.__trigger_metadata is None:
            return None

        if self._trigger_metadata_pyobj is None:
            # No need to do deepcopy since datum.python_value will construct
            # new object
            self._trigger_metadata_pyobj = {
                k: v.python_value for (k, v) in self.__trigger_metadata.items()
            }
        return self._trigger_metadata_pyobj

    def __repr__(self) -> str:
        return (
            f'<azure.functions.ServiceBusMessage '
            f'message_id={self.message_id} '
            f'at 0x{id(self):0x}>'
        )


class ServiceBusMessageInConverter(meta.InConverter,
                                   binding='serviceBusTrigger', trigger=True):

    @classmethod
    def check_input_type_annotation(cls, pytype: type) -> bool:
        valid_types = (azf_sbus.ServiceBusMessage)
        return (
            meta.is_iterable_type_annotation(pytype, valid_types)
            or (isinstance(pytype, type) and issubclass(pytype, valid_types))
        )

    @classmethod
    def decode(
        cls, data: meta.Datum, *, trigger_metadata: Mapping[str, meta.Datum]
    ) -> Union[ServiceBusMessage, List[ServiceBusMessage]]:
        """Returns the application setting from environment variable.

        Parameters
        ----------
        data: meta.Datum
            The datum from GRPC message

        trigger_metadata: Mapping[str, meta.Datum]
            The metadata of the Service Bus trigger, usually populated by
            function host

        Returns
        -------
        Union[ServiceBusMessage, List[ServiceBusMessage]]
            When 'cardinality' is set to 'one', this method returns a single
            ServiceBusMessage. When 'cardinality' is set to 'many' this method
            returns a list of ServiceBusMessage.
        """
        if cls._is_cardinality_one(trigger_metadata):
            return cls.decode_single_message(
                data, trigger_metadata=trigger_metadata)
        elif cls._is_cardinality_many(trigger_metadata):
            return cls.decode_multiple_messages(
                data, trigger_metadata=trigger_metadata)
        else:
            raise NotImplementedError(
                f'unsupported service bus data type: {data.type} or '
                'UserProperties does not exist')

    @classmethod
    def decode_single_message(
        cls, data: meta.Datum, *,
        trigger_metadata: Mapping[str, meta.Datum]
    ) -> ServiceBusMessage:
        if data is None:
            # ServiceBus message with no payload are possible.
            # See Azure/azure-functions-python-worker#330
            body = b''

        elif data.type in ['string', 'json']:
            body = data.value.encode('utf-8')

        elif data.type == 'bytes':
            body = data.value

        else:
            raise NotImplementedError(
                f'unsupported queue payload type: {data.type}')

        if trigger_metadata is None:
            raise NotImplementedError(
                'missing trigger metadata for ServiceBus message input')

        return ServiceBusMessage(
            body=body,
            trigger_metadata=trigger_metadata,
            content_type=cls._decode_trigger_metadata_field(
                trigger_metadata, 'ContentType', python_type=str),
            correlation_id=cls._decode_trigger_metadata_field(
                trigger_metadata, 'CorrelationId', python_type=str),
            dead_letter_source=cls._decode_trigger_metadata_field(
                trigger_metadata, 'DeadLetterSource', python_type=str),
            delivery_count=cls._decode_trigger_metadata_field(
                trigger_metadata, 'DeliveryCount', python_type=int),
            enqueued_time_utc=cls._parse_datetime_metadata(
                trigger_metadata, 'EnqueuedTimeUtc'),
            expires_at_utc=cls._parse_datetime_metadata(
                trigger_metadata, 'ExpiresAtUtc'),
            label=cls._decode_trigger_metadata_field(
                trigger_metadata, 'Label', python_type=str),
            lock_token=cls._decode_trigger_metadata_field(
                trigger_metadata, 'LockToken', python_type=str),
            message_id=cls._decode_trigger_metadata_field(
                trigger_metadata, 'MessageId', python_type=str),
            partition_key=cls._decode_trigger_metadata_field(
                trigger_metadata, 'PartitionKey', python_type=str),
            reply_to=cls._decode_trigger_metadata_field(
                trigger_metadata, 'ReplyTo', python_type=str),
            reply_to_session_id=cls._decode_trigger_metadata_field(
                trigger_metadata, 'ReplyToSessionId', python_type=str),
            scheduled_enqueue_time_utc=cls._parse_datetime_metadata(
                trigger_metadata, 'ScheduledEnqueueTimeUtc'),
            sequence_number=cls._decode_trigger_metadata_field(
                trigger_metadata, 'SequenceNumber', python_type=int),
            session_id=cls._decode_trigger_metadata_field(
                trigger_metadata, 'SessionId', python_type=str),
            time_to_live=cls._parse_timedelta_metadata(
                trigger_metadata, 'TimeToLive'),
            to=cls._decode_trigger_metadata_field(
                trigger_metadata, 'To', python_type=str),
            user_properties=cls._decode_trigger_metadata_field(
                trigger_metadata, 'UserProperties', python_type=dict),
        )

    @classmethod
    def decode_multiple_messages(
        cls, data: meta.Datum, *,
        trigger_metadata: Mapping[str, meta.Datum]
    ) -> List[ServiceBusMessage]:
        """Unlike EventHub, the trigger_metadata already contains a set of
        arrays (e.g. 'ContentTypeArray', 'CorrelationidArray'...). We can
        retrieve message properties directly from those array.
        """
        if data.type == 'collection_bytes':
            parsed_data = data.value.bytes

        elif data.type == 'collection_string':
            parsed_data = data.value.string

        # Input Trigger IotHub Event
        elif data.type == 'json':
            parsed_data = json.loads(data.value)

        else:
            raise NotImplementedError('unable to decode multiple messages '
                                      f'with data type {data.type}')

        return cls._extract_messages(parsed_data, data.type, trigger_metadata)

    @classmethod
    def _is_cardinality_many(
        cls,
        trigger_metadata: Mapping[str, meta.Datum]
    ) -> bool:
        return 'UserPropertiesArray' in trigger_metadata

    @classmethod
    def _is_cardinality_one(
        cls,
        trigger_metadata: Mapping[str, meta.Datum]
    ) -> bool:
        return 'UserProperties' in trigger_metadata

    @classmethod
    def _get_event_count(
        cls,
        trigger_metadata: Mapping[str, meta.Datum]
    ) -> int:
        datum = trigger_metadata['UserPropertiesArray']
        user_props = json.loads(datum.value)
        return len(user_props)

    @classmethod
    def _marshall_message_body(
        cls,
        body: Union[bytes, str, Dict[str, Any]],
        data_type: str
    ) -> bytes:
        if data_type == 'bytes' and isinstance(body, bytes):
            return body
        elif data_type == 'str' and isinstance(body, str):
            return body.encode('utf-8')
        elif data_type == 'json' and isinstance(body, dict):
            return json.dumps(body).encode('utf-8')
        else:
            raise NotImplementedError('unable to marshall message body with '
                                      f'data_type {data_type}')

    @classmethod
    def _marshall_message_bodies(
        cls,
        bodies: Union[List[str], List[bytes]],
        data_type: str
    ) -> List[bytes]:
        # The typing library cast() method is used as mypy type helper
        # Currently, mypy does not provide List[type] checking for now
        # Thus, forcefully casting is required
        if data_type == 'collection_bytes':
            return cast(List[bytes], bodies)
        elif data_type == 'collection_string':
            strings: List[str] = cast(List[str], bodies)
            return cast(List[bytes], [b.encode('utf-8') for b in strings])
        elif data_type == 'json':
            return cast(List[bytes],
                        [json.dumps(b).encode('utf-8') for b in bodies])
        else:
            raise NotImplementedError('unable to marshall message bodies with '
                                      f'data_type {data_type}')

    @classmethod
    def _extract_messages(
        cls, parsed_data: Union[List[bytes], List[str]], data_type: str,
        trigger_metadata: Mapping[str, meta.Datum]
    ) -> List[ServiceBusMessage]:
        messages: List[ServiceBusMessage] = []
        num_messages: int = cls._get_event_count(trigger_metadata)
        message_bodies: List[bytes] = cls._marshall_message_bodies(
            bodies=parsed_data, data_type=data_type
        )
        for i in range(num_messages):
            messages.append(ServiceBusMessage(
                body=message_bodies[i],
                trigger_metadata=trigger_metadata,
                content_type=cls._get_from_metadata_array(
                    trigger_metadata, 'ContentTypeArray', i),
                correlation_id=cls._get_from_metadata_array(
                    trigger_metadata, 'CorrelationIdArray', i),
                dead_letter_source=cls._get_from_metadata_array(
                    trigger_metadata, 'DeadLetterSourceArray', i),
                delivery_count=cls._get_from_metadata_array(
                    trigger_metadata, 'DeliveryCountArray', i),
                enqueued_time_utc=cls._parse_datetime(
                    cls._get_from_metadata_array(
                        trigger_metadata, 'EnqueuedTimeUtcArray', i)),
                expires_at_utc=cls._parse_datetime(
                    cls._get_from_metadata_array(
                        trigger_metadata, 'ExpiresAtUtcArray', i)),
                label=cls._get_from_metadata_array(
                    trigger_metadata, 'LabelArray', i),
                lock_token=cls._get_from_metadata_array(
                    trigger_metadata, 'LockTokenArray', i),
                message_id=cls._get_from_metadata_array(
                    trigger_metadata, 'MessageIdArray', i),
                partition_key=cls._get_from_metadata_array(
                    trigger_metadata, 'PartitionKeyArray', i),
                reply_to_session_id=cls._get_from_metadata_array(
                    trigger_metadata, 'ReplyToSessionIdArray', i),
                scheduled_enqueue_time_utc=cls._parse_datetime(
                    cls._get_from_metadata_array(
                        trigger_metadata, 'ScheduledEnqueueTimeUtcArray', i)),
                sequence_number=cls._get_from_metadata_array(
                    trigger_metadata, 'SequenceNumberArray', i),
                session_id=cls._get_from_metadata_array(
                    trigger_metadata, 'SessionIdArray', i),
                time_to_live=cls._parse_timedelta(
                    cls._get_from_metadata_array(
                        trigger_metadata, 'TimeToLiveArray', i)),
                to=cls._get_from_metadata_array(
                    trigger_metadata, 'ToArray', i),
                reply_to=cls._get_from_metadata_array(
                    trigger_metadata, 'ReplyToArray', i),
                user_properties=cls._get_from_metadata_array(
                    trigger_metadata, 'UserPropertiesArray', i)
            ))
        return messages

    @classmethod
    def _get_from_metadata_array(
        cls,
        trigger_metadata: Mapping[str, meta.Datum],
        array_name: str,
        index: int
    ) -> Any:
        """This method is to safe-guard when retrieve data from arrays.

        Some array may be missing in metadata. Others may not contain certain
        values (e.g. correlation_ids array may be empty [] when there's
        multiple messages).

        Parameters
        ----------
        trigger_metadata: Mapping[str, meta.Datum]
            The trigger metadata that contains multiple ServiceBus messages
        array_name: str
            The name of the array needs to be extracted
        index: int
            The element index wants to be extracted

        Returns
        -------
        Optional[Any]
            If the array name does not exist in trigger_metadata, returns None.
            If the array does not contains certain element, or index out of
            bound, returns None.
            Otherwise, return the element.
        """

        # Check if array name does exist (e.g. ContentTypeArray)
        datum: Optional[meta.Datum] = trigger_metadata.get(array_name)
        if datum is None:
            return None

        # Check if datum is an iterable element (e.g. collection_string)
        data_array: Optional[Union[List[str], List[int], List[bytes]]] = None
        if datum.type == 'collection_string':
            data_array = datum.value.string
        elif datum.type == 'collection_bytes':
            data_array = datum.value.bytes
        elif datum.type == 'collection_sint64':
            data_array = datum.value.sint64
        elif datum.type == 'json':
            data_array = json.loads(datum.value)

        # Check if the index is inbound
        if data_array is None or index >= len(data_array):
            return None

        return data_array[index]


class ServiceBusMessageOutConverter(meta.OutConverter, binding='serviceBus'):

    @classmethod
    def check_output_type_annotation(cls, pytype: type) -> bool:
        return issubclass(pytype, (str, bytes))

    @classmethod
    def encode(cls, obj: Any, *,
               expected_type: Optional[type]) -> meta.Datum:
        if isinstance(obj, str):
            return meta.Datum(type='string', value=obj)

        elif isinstance(obj, bytes):
            return meta.Datum(type='bytes', value=obj)

        raise NotImplementedError



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

from typing import Dict, List, Tuple, Optional, Any, Union
import logging
import asyncio
from asyncio import Event, Queue
from warnings import warn
from wsgiref.headers import Headers

from ._abc import Context
from ._http import HttpRequest, HttpResponse
from ._http_wsgi import WsgiRequest

ASGI_VERSION = "2.1"
ASGI_SPEC_VERSION = "2.1"


class AsgiRequest(WsgiRequest):
    def __init__(self, func_req: HttpRequest,
                 func_ctx: Optional[Context] = None):
        self.asgi_version = ASGI_VERSION
        self.asgi_spec_version = ASGI_SPEC_VERSION
        self._headers = func_req.headers
        super().__init__(func_req, func_ctx)

    def _get_encoded_http_headers(self) -> List[Tuple[bytes, bytes]]:
        return [(k.encode("utf8"), v.encode("utf8"))
                for k, v in self._headers.items()]

    def _get_server_address(self):
        if self.server_name is not None:
            return (self.server_name, int(self.server_port))
        return None

    def to_asgi_http_scope(self):
        if self.path_info is not None:
            _raw_path = self.path_info.encode("utf-8")
        else:
            _raw_path = b''
        if self.query_string is not None:
            _query_string = self.query_string.encode("utf-8")
        else:
            _query_string = b''

        return {
            "type": "http",
            "asgi.version": self.asgi_version,
            "asgi.spec_version": self.asgi_spec_version,
            "http_version": "1.1",
            "method": self.request_method,
            "scheme": "https",
            "path": self.path_info,
            "raw_path": _raw_path,
            "query_string": _query_string,
            "root_path": self.script_name,
            "headers": self._get_encoded_http_headers(),
            "server": self._get_server_address(),
            "client": None,
            "azure_functions.function_directory": self.af_function_directory,
            "azure_functions.function_name": self.af_function_name,
            "azure_functions.invocation_id": self.af_invocation_id,
            "azure_functions.thread_local_storage":
                self.af_thread_local_storage,
            "azure_functions.trace_context": self.af_trace_context,
            "azure_functions.retry_context": self.af_retry_context
        }
        # Notes, missing client name, port


class AsgiResponse:
    def __init__(self):
        self._status_code = 0
        self._headers: Union[Headers, Dict] = {}
        self._buffer: List[bytes] = []
        self._request_body: Optional[bytes] = b""
        self._has_received_response: bool = False

    @classmethod
    async def from_app(cls, app, scope: Dict[str, Any],
                       body: bytes) -> "AsgiResponse":
        res = cls()
        res._request_body = body
        await app(scope, res._receive, res._send)
        return res

    def to_func_response(self) -> HttpResponse:
        lowercased_headers = {k.lower(): v for k, v in self._headers.items()}
        return HttpResponse(
            body=b"".join(self._buffer),
            status_code=self._status_code,
            headers=self._headers,  # type: ignore
            mimetype=lowercased_headers.get("content-type"),
            charset=lowercased_headers.get("content-encoding"),
        )

    def _handle_http_response_start(self, message: Dict[str, Any]):
        self._headers = Headers(
            [(k.decode(), v.decode())
             for k, v in message["headers"]])
        self._status_code = message["status"]

    def _handle_http_response_body(self, message: Dict[str, Any]):
        self._buffer.append(message["body"])
        self._has_received_response = not message.get("more_body", False)
        # XXX : Chunked bodies not supported, see
        # https://github.com/Azure/azure-functions-host/issues/4926

    async def _receive(self):
        if self._request_body is not None:
            reply = {
                "type": "http.request",
                "body": self._request_body,
                "more_body": False,
            }
            self._request_body = None
            return reply
        else:
            while not self._has_received_response:
                await asyncio.sleep(0.1)
            return {
                "type": "http.disconnect",
            }

    async def _send(self, message):
        logging.debug("Received %s from ASGI worker.", message)
        if message["type"] == "http.response.start":
            self._handle_http_response_start(message)
        elif message["type"] == "http.response.body":
            self._handle_http_response_body(message)
        elif message["type"] == "http.disconnect":
            pass  # Nothing todo here


class AsgiMiddleware:
    """This middleware is to adapt an ASGI supported Python server
    framework into Azure Functions. It can be used by either calling the
    .handle() function or exposing the .main property in a HttpTrigger.
    """
    _logger = logging.getLogger('azure.functions.AsgiMiddleware')
    _usage_reported = False

    def __init__(self, app):
        """Instantiate an ASGI middleware to convert Azure Functions HTTP
        request into ASGI Python object. Example on handling ASGI app in a HTTP
        trigger by overwriting the .main() method:

        import azure.functions as func

        from FastapiApp import app

        main = func.AsgiMiddleware(app).main
        """
        if not self._usage_reported:
            self._logger.debug("Starting Azure Functions ASGI middleware.")
            self._usage_reported = True

        self._app = app
        self.main = self._handle
        self.state = {}
        self.lifespan_receive_queue: Optional[Queue] = None
        self.lifespan_startup_event: Optional[Event] = None
        self.lifespan_shutdown_event: Optional[Event] = None
        self._startup_succeeded = False

    def handle(self, req: HttpRequest, context: Optional[Context] = None):
        """Deprecated. Please use handle_async instead:

        import azure.functions as func

        from FastapiApp import app

        async def main(req, context):
            return await func.AsgiMiddleware(app).handle_async(req, context)
        """
        warn("handle() is deprecated. Please await .handle_async() instead.",
             DeprecationWarning, stacklevel=2)
        self._logger.warning(
            "handle() is deprecated. Please `await .handle_async()` instead.")
        return self._handle(req, context)

    def _handle(self, req, context):
        asgi_request = AsgiRequest(req, context)
        scope = asgi_request.to_asgi_http_scope()
        asgi_response = asyncio.run(
            AsgiResponse.from_app(self._app, scope, req.get_body())
        )

        return asgi_response.to_func_response()

    async def handle_async(self,
                           req: HttpRequest,
                           context: Optional[Context] = None):
        """Method to convert an Azure Functions HTTP request into a ASGI
        Python object. Example on handling ASGI app in a HTTP trigger by
        calling .handle_async() in .main() method:

            import azure.functions as func

            from FastapiApp import app

            async def main(req, context):
                return await func.AsgiMiddleware(app).handle_async(req,
                                                                   context)
        """
        return await self._handle_async(req, context)

    async def _handle_async(self, req, context):
        asgi_request = AsgiRequest(req, context)
        scope = asgi_request.to_asgi_http_scope()
        asgi_response = await AsgiResponse.from_app(self._app,
                                                    scope,
                                                    req.get_body())
        return asgi_response.to_func_response()

    async def _lifespan_receive(self):
        if not self.lifespan_receive_queue:
            raise RuntimeError("notify_startup() must be called first.")
        return await self.lifespan_receive_queue.get()

    async def _lifespan_send(self, message):
        logging.debug("Received lifespan message %s.", message)
        if not self.lifespan_startup_event or not self.lifespan_shutdown_event:
            raise RuntimeError("notify_startup() must be called first.")
        if message["type"] == "lifespan.startup.complete":
            self.lifespan_startup_event.set()
            self._startup_succeeded = True
        elif message["type"] == "lifespan.shutdown.complete":
            self.lifespan_shutdown_event.set()
        elif message["type"] == "lifespan.startup.failed":
            self.lifespan_startup_event.set()
            self._startup_succeeded = False
            if message.get("message"):
                self._logger.error("Failed ASGI startup with message '%s'.",
                                   message["message"])
            else:
                self._logger.error("Failed ASGI startup event.")
        elif message["type"] == "lifespan.shutdown.failed":
            self.lifespan_shutdown_event.set()
            if message.get("message"):
                self._logger.error("Failed ASGI shutdown with message '%s'.",
                                   message["message"])
            else:
                self._logger.error("Failed ASGI shutdown event.")

    async def _lifespan_main(self):
        scope = {
            "type": "lifespan",
            "asgi.version": ASGI_VERSION,
            "asgi.spec_version": ASGI_SPEC_VERSION,
            "state": self.state,
        }
        if not self.lifespan_startup_event or not self.lifespan_shutdown_event:
            raise RuntimeError("notify_startup() must be called first.")
        try:
            await self._app(scope, self._lifespan_receive, self._lifespan_send)
        finally:
            self.lifespan_startup_event.set()
            self.lifespan_shutdown_event.set()

    async def notify_startup(self):
        """Notify the ASGI app that the server has started."""
        self._logger.debug("Notifying ASGI app of startup.")

        # Initialize signals and queues
        if not self.lifespan_receive_queue:
            self.lifespan_receive_queue = Queue()
        if not self.lifespan_startup_event:
            self.lifespan_startup_event = Event()
        if not self.lifespan_shutdown_event:
            self.lifespan_shutdown_event = Event()

        startup_event = {"type": "lifespan.startup"}
        await self.lifespan_receive_queue.put(startup_event)
        task = asyncio.create_task(self._lifespan_main())  # NOQA
        await self.lifespan_startup_event.wait()
        return self._startup_succeeded

    async def notify_shutdown(self):
        """Notify the ASGI app that the server is shutting down."""
        if not self.lifespan_receive_queue or not self.lifespan_shutdown_event:
            raise RuntimeError("notify_startup() must be called first.")

        self._logger.debug("Notifying ASGI app of shutdown.")
        shutdown_event = {"type": "lifespan.shutdown"}
        await self.lifespan_receive_queue.put(shutdown_event)
        await self.lifespan_shutdown_event.wait()



#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License.
from typing import Optional

from azure.functions.decorators.constants import WARMUP_TRIGGER
from azure.functions.decorators.core import Trigger, DataType


class WarmUpTrigger(Trigger):
    @staticmethod
    def get_binding_name() -> str:
        return WARMUP_TRIGGER

    def __init__(self,
                 name: str,
                 data_type: Optional[DataType] = None,
                 **kwargs) -> None:
        super().__init__(name=name, data_type=data_type)



#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License.
import unittest

from azure.functions.decorators.function_name import FunctionName


class TestFunctionName(unittest.TestCase):

    def test_retry_policy_setting_creation(self):
        function_name = FunctionName(function_name="TestFunctionName")

        self.assertEqual(function_name.get_setting_name(), "function_name")
        self.assertEqual(function_name.get_dict_repr(),
                         {'setting_name': 'function_name',
                          'function_name': 'TestFunctionName'})
